/* Stress floating point operations across a specieid amount of memo */

#include <unistd.h>
#include "fstress.h"

double **wspace ;     /* one array per thread */
double *thrdres ;     /* result from each thread */
unsigned int nelem ;  /* number of elements per thread */
unsigned int niter ;  /* number of iterations per thread */

FILE *flog;

int main(int argc, char *argv[])
{
	int memsize = 4;  /* tot mem size per thread in MB */
	int nthreads = 1;  /* number of threads */
	struct doit_args *thread_ids;
	pthread_t *threadbufs;
	pthread_barrier_t barrier;
	double thrdsum;
	int thrd;

	flog = fopen("fstress1.pid", "w");
	if (!flog) {
		printf("Can't open pidfile: %m\n");
		exit(1);
	}
	fprintf(flog, "%d", getpid());
	fclose(flog);

	flog = fopen("fstrees.log", "w");
	if (!flog) {
		printf("Can't open log: %m\n");
		exit(1);
	}

	niter = 10 ;

	if  (argc == 2 && strcmp(argv[1],"-h") == 0)  {
		printf("fstress memsiz(MB)[%d] niterations[%d] nthreads[%d]\n",
		       memsize, niter, nthreads) ;
		exit(0) ;
	}

	if  (argc >= 2)  {
		memsize = atoi(argv[1]) ;
		if  (memsize < 1 || memsize > 10000)
			ERR2("bad number of MB: %d", memsize);
	}

	if  (argc >= 3)  {
		niter = atoi(argv[2]) ;
		if  (niter < 1 || niter > 1000000000)
			ERR2("bad number of iterations: %d", niter);
	}

	if  (argc >= 4)  {
		nthreads = atoi(argv[3]) ;
		if  (nthreads < 1 || nthreads > 100)
			ERR2("bad number of threads: %d", nthreads);
	}

	pr_info("fstress %d %d %d\n\n", memsize, niter, nthreads) ;

	pthread_barrier_init(&barrier, NULL, nthreads + 1);

	wspace = calloc(1, sizeof(double *) * nthreads);
	if  (!wspace)
		ERRperror("calloc pointer to workspaces");
	pr_info("wspace %p\n", wspace);

	nelem = memsize * 1000000 / sizeof(double);
	for  (thrd = 0 ; thrd < nthreads ; thrd++)  {
		wspace[thrd] = calloc(1, sizeof(double) * nelem);
		if  (!wspace[thrd])
			ERRperror("calloc workspaces");
		pr_info("wspace[%d]=%p\n", thrd, wspace[thrd]);
	}

	thrdres = calloc(1, sizeof(double) * nthreads);
	if  (!thrdres)
		ERRperror("calloc thrdres");
	pr_info("thrdres %p\n", thrdres);

	populate(wspace, nelem, nthreads) ;

	threadbufs = calloc(1, sizeof(pthread_t) * nthreads);
	if  (!threadbufs)
		ERRperror("calloc thread contexts");
	pr_info("threadbufs %p\n", threadbufs);

	if  ((thread_ids = calloc(1, sizeof(*thread_ids) * nthreads)) == NULL)
		ERRperror("calloc thread ids");
	pr_info("thread_ids %p\n", thread_ids);

#if 1
	for  (thrd = 0 ; thrd < nthreads ; thrd++)  {
		thread_ids[thrd].thrd_num = thrd ;
		thread_ids[thrd].barrier = &barrier ;
		if  (pthread_create(&threadbufs[thrd], NULL,
				    doit, (void *)&thread_ids[thrd]) != 0)
			ERRperror("pthread_create") ;
	}

	pthread_barrier_wait(&barrier);

	thrdsum = (double)0.0 ;
	for  (thrd = 0 ; thrd < nthreads ; thrd++)  {
		pthread_join(threadbufs[thrd], NULL) ;
		thrdsum = thrdsum + thrdres[thrd] ;
	}
#else
	for  (thrd = 0 ; thrd < nthreads ; thrd++)  {
		thread_ids[thrd] = thrd ;
		doit(&thread_ids[thrd]);
	}

	thrdsum = (double)0.0 ;
	for  (thrd = 0 ; thrd < nthreads ; thrd++)  {
		thrdsum = thrdsum + thrdres[thrd] ;
	}
#endif

	pr_info("total: %20.15f\n",thrdsum) ;
	for  (thrd = 0 ; thrd < nthreads ; thrd++)  {
		pr_info("%15.12f ", thrdres[thrd]) ;
	}
	pr_info("\n") ;
}
